// inetPort.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/*
    Author: Marty Kraimer
    Date: 23JAN2003

*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include <osiSock.h>
#include <iocsh.h>
#include <epicsThread.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsTime.h>
#include <epicsAssert.h>
#include <iocsh.h>
#include <mpfType.h>
#include <WatchDog.h>
#include "SerialPort.h"
#include <epicsExport.h>

enum tcpConnectState {tcpConnectYes, tcpConnectNo};
typedef void(*tcpConnectCallback)(void *pvt,tcpConnectState state);
typedef void(*tcpReceiveCallback)(void *pvt,void *buffer,int32 nbytes);

class InetTcp
{
public:
    InetTcp(int port,const char *address,int bufferSize);
    ~InetTcp();
    void connect(void *userPvt, tcpConnectCallback ,tcpReceiveCallback);
    int sendFlush(void *data,int32 size);
private:
    void *tcpPvt;
    // disallow following
    InetTcp();
    InetTcp(const InetTcp &);
    InetTcp& operator=(const InetTcp&);
};

class InetPort : public SerialPort{
public:
    InetPort(char *portName,char *inetAddr,int port,
        InetTcp *pInetTcp,int bufsize);
    virtual serialStatus write(
	unsigned char *data, int nbytes, double timeout);
    virtual serialStatus read(double timeout);
    virtual void flush();
    char *portName;
    char *inetAddr;
    int port;
    InetTcp *pInetTcp;
    tcpConnectState inetState;
    int bufsize;
    int nextFree;
    char *buffer;
    epicsEventId dataArrived;
    epicsMutexId lock;
};

extern "C" {
int inettcpDebug=0;
epicsExportAddress(int,inettcpDebug);
static const int nConnectTry = 5;
static const double connectRetryDelay = 10;

int initInetPort(char *portName, char *inetAddr,int port,int bufsize);
} // extern "C"

int initInetPort(char *portn,char *inet,int port,int bufsize)
{
    char *portName,*inetAddr;
    portName = (char *)calloc(strlen(portn)+1,sizeof(char));
    strcpy(portName,portn);
    inetAddr = (char *)calloc(strlen(inet)+1,sizeof(char));
    strcpy(inetAddr,inet);
    InetTcp *pInetTcp = new InetTcp(port,inetAddr,bufsize);
    InetPort *pInetPort = new InetPort(portName,inetAddr,port,pInetTcp,bufsize);
    pInetPort->registerName(portName);
    return(0);
}

static void connectCallback(void *pvt,tcpConnectState state)
{
    InetPort *pInetPort = (InetPort *)pvt;
    epicsMutexMustLock(pInetPort->lock);
    pInetPort->inetState = state;
    epicsMutexUnlock(pInetPort->lock);
}

static void receiveCallback(void *pvt,void *buffer,int32 nbytes)
{
    InetPort *pInetPort = (InetPort *)pvt;
    epicsMutexMustLock(pInetPort->lock);
    if(nbytes > (pInetPort->bufsize - pInetPort->nextFree)) {
        printf("InetPort %s %s %d : received %d but bufsize %d nextFree %d\n",
            pInetPort->portName,pInetPort->inetAddr,pInetPort->port,
            nbytes,pInetPort->bufsize,pInetPort->nextFree);
        nbytes = pInetPort->bufsize - pInetPort->nextFree;
    }
    memcpy(pInetPort->buffer +pInetPort->nextFree,buffer,nbytes);
    pInetPort->nextFree += nbytes;
    epicsMutexUnlock(pInetPort->lock);
    epicsEventSignal(pInetPort->dataArrived);
}

InetPort::InetPort(char *port,char *inet, int p,
    InetTcp *ptcp,int size)
: portName(port), inetAddr(inet), port(p),
  pInetTcp(ptcp), inetState(tcpConnectNo), bufsize(size), nextFree(0)
{
    buffer = new char[bufsize];
    dataArrived = epicsEventMustCreate(epicsEventEmpty);
    lock = epicsMutexMustCreate();
    pInetTcp->connect(this,connectCallback,receiveCallback);
}

serialStatus InetPort::write(
    unsigned char *data, int nbytes, double )
{
    if(inetState!=tcpConnectYes)  return(serialFailure);
    int status = pInetTcp->sendFlush(data,nbytes);
    if(status) return(serialFailure);
    if(peekHandler) {
        for(int ind=0; ind<nbytes; ind++) {
            (*peekHandler)(peekHandlerParm,false,data[ind]);
        }
    }
    return(serialSuccess);
}

serialStatus InetPort::read(double timeOut)
{
    byteHandlerRC rc = byteHandlerError;
    double timeLeft = timeOut;
    epicsTimeStamp start,now;

    assert(byteHandler);
    if(inetState!=tcpConnectYes)  return(serialFailure);
    epicsTimeGetCurrent(&start);
    while(1) {
        epicsEventWaitStatus status;
        status = epicsEventWaitWithTimeout(dataArrived,timeLeft);
        if(status!=epicsEventWaitOK) {
            return((status==epicsEventWaitTimeout)
                   ? serialTimeout : serialFailure);
        }
        epicsMutexMustLock(lock);
        for(int ind=0; ind<nextFree; ind++) {
            if(peekHandler) (*peekHandler)
                (peekHandlerParm,true,(unsigned char)buffer[ind]);
            rc = (*byteHandler)(byteHandlerParm,(unsigned char)buffer[ind]);
            if(rc!=byteHandlerOK) {
                if(rc==byteHandlerEndRead) {
                    int nextNotUsed = ind + 1;
                    if(nextNotUsed<nextFree) {
                        int nleft = nextFree - nextNotUsed + 1;
                        memmove(buffer,&buffer[nextNotUsed],nleft);
                        nextFree = nleft;
                    } else {
                        nextFree = 0;
                    }
                }
                serialStatus ioStatus = (rc==byteHandlerEndRead)
                    ? serialSuccess : serialFailure;
                epicsMutexUnlock(lock);
                return(ioStatus);
            }
        }
        nextFree = 0;
        epicsMutexUnlock(lock);
        epicsTimeGetCurrent(&now);
        timeLeft = epicsTimeDiffInSeconds(&now,&start);
    }
}

void InetPort::flush()
{
    nextFree = 0;
}
/* tcp code */

//The following is for exclusive use by this source module
// InetTcp is an interface class that uses InetTcpPrivate for implementation
class InetTcpPrivate {
public:
    InetTcpPrivate(int port,const char *address,int bufferSize);
    ~InetTcpPrivate();
    void connect(void *userPvt, tcpConnectCallback ,tcpReceiveCallback);
    int sendFlush(void *data,int32 size);
    //following are for use by this source module
    void lock() {epicsMutexMustLock(lockSem);}
    void unlock() {epicsMutexUnlock(lockSem);}
    void issueError(const char *message);
    void issueErrorExit(const char *message);
    void dumpData(const char *who,void *data, int32 nbytes);
    int initSocket();
    int connectServer();
    int connectClient();
    //beginning of data
    tcpConnectCallback connectCallback;
    tcpReceiveCallback receiveCallback;
    void *userPvt;
    int port;
    const int bufferSize;
    bool connected;
    epicsThreadId threadid;
    char *address;
    char *inbuf;
    int sockfd;
    struct sockaddr_in servaddr;
    struct sockaddr_in clientaddr; // Used by server only
    epicsMutexId lockSem;
private:
    //disallow following
    InetTcpPrivate();
    InetTcpPrivate(const InetTcpPrivate &);
    InetTcpPrivate & operator=(const InetTcpPrivate &);
};

extern "C" {
void inettcpTask(InetTcpPrivate *ptcp);
}

InetTcp::InetTcp(int port,const char *address,int bufferSize)
{ tcpPvt = new InetTcpPrivate(port,address,bufferSize);}

InetTcp::~InetTcp() {delete (InetTcpPrivate *)tcpPvt;}

void InetTcp::connect(void *userPvt, tcpConnectCallback connectCallback,
    tcpReceiveCallback receiveCallback)
{
    InetTcpPrivate *pvt = (InetTcpPrivate *)tcpPvt;
    pvt->connect(userPvt,connectCallback,receiveCallback);
    return;
}

int InetTcp::sendFlush(void *data,int32 size)
{
    InetTcpPrivate *pvt = (InetTcpPrivate *)tcpPvt;
    pvt->lock();
    int status = pvt->sendFlush(data,size);
    pvt->unlock();
    return(status);
}

InetTcpPrivate::InetTcpPrivate(int port,const char *addr,int bufsize)
    : connectCallback(0), receiveCallback(0), userPvt(0),
    port(port), bufferSize(bufsize), connected(false)
{
    if(bufferSize<=10) {
	printf("InetTcpPrivate::InetTcpPrivate illegal bufferSize %d\n",bufferSize);
	return;
    }
    lockSem = epicsMutexMustCreate();
    address = new char[::strlen(addr)+1];
    inbuf = new char[bufferSize];
    ::strcpy(address,addr);
}

InetTcpPrivate::~InetTcpPrivate()
{
    if(sockfd>0) {
	int status = ::close(sockfd);
	if(status) issueError("closing sockfd");
    }
    delete [] inbuf;
    delete [] address;
    epicsMutexDestroy(lockSem);
}

static const char *InetTcpClient = "InetTcpClient";
static const char *InetTcpServer = "InetTcpServer";

void InetTcpPrivate::connect(void *pvt,
tcpConnectCallback connectcallback,tcpReceiveCallback receivecallback)
{
    char *name;
    userPvt = pvt;
    connectCallback = connectcallback;
    receiveCallback = receivecallback;
    name = new char[::strlen(InetTcpClient) + 1];
    ::strcpy(name,InetTcpClient);
    threadid = ::epicsThreadCreate(name,
        epicsThreadPriorityScanHigh+2,
        epicsThreadGetStackSize(epicsThreadStackMedium),
	(EPICSTHREADFUNC)inettcpTask, (void *)this);
    if(threadid==0) {
	issueErrorExit("InetTcp::connect epicsThreadCreate error");
    }
}

int InetTcpPrivate::sendFlush(void *data,int32 size)
{
    if(inettcpDebug) dumpData("InetTcp::sendFlush:",data,size);
    if(!connected) return(-1);
    int nbytes = ::send(sockfd,(char *)data,size,0);
    if(nbytes != size) {
        printf("InetTcpPrivate::sendFlush (%d,%d) bytes (requested,sent)\n",
            size,nbytes);
        return(-1);
    }
    return(0);
}

void InetTcpPrivate::issueError(const char *message)
{
    fprintf(stderr,"InetTcp error: %s errno %s\n",message,strerror(errno));
}

void InetTcpPrivate::issueErrorExit(const char *message)
{
    issueError(message);
    epicsThreadSuspendSelf();
}


void InetTcpPrivate::dumpData(const char *who,void *data, int32 nbytes)
{
    printf("%s",who);
    printf(" nbytes %d ",nbytes);
    char *p = (char *)data;
    for(int i=0; i<nbytes; i++) {
       printf("%c %2.2x ",*p,*p);
       ++p;
    }
    printf("\n");
}

int InetTcpPrivate::initSocket()
{
    sockfd = epicsSocketCreate(AF_INET, SOCK_STREAM, 0);
    if(sockfd < 0) {
	issueError("InetTcp: socket error");
	return(-1);
    }
    int trueValue = 1;
    int status = ::setsockopt(sockfd,IPPROTO_TCP,TCP_NODELAY,
	(char *)&trueValue, sizeof(trueValue));
    if(status!=0) {
	issueError("InetTcp: setsockopt error");
	return -1;
    }
    int size = bufferSize;
    status = ::setsockopt(sockfd,SOL_SOCKET,SO_SNDBUF,
	(char *)&size,sizeof(size));
    if(status!=0) {
	issueError("InetTcp: setsockopt error");
	return -1;
    }
    status = ::setsockopt(sockfd,SOL_SOCKET,SO_RCVBUF,
	(char *)&size,sizeof(size));
    if(status!=0) {
	issueError("InetTcp: setsockopt error");
	return -1;
    }
    memset(&servaddr,0,sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    if(address) {
        status = ::aToIPAddr(address,port,&servaddr);
        if(status) {
            issueError("InetTcp::initSocket aToIPAddr failed\n");
            return(-1);
        }
    } else {
printf("InetTcpPrivate::initSocket why are we here?\n");
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
        servaddr.sin_port = htons(port);
    }
    return(0);
}

int InetTcpPrivate::connectClient()
{
    int rtn;
    bool firstTime = true;
    while(true) {
        if(initSocket()) return(-1);
	rtn = ::connect(sockfd,(struct sockaddr *)&servaddr,sizeof(servaddr));
	if(rtn==0) break;
	if(firstTime) {
	    issueError("InetTcpPrivate::connectClient connect error");
	    firstTime = false;
	}
	rtn = ::close(sockfd);
	if(rtn) issueError("cpPrivate::connectClient close error");
	epicsThreadSleep(240);
    }
    connected = true;
    return(0);
}

void inettcpTask(InetTcpPrivate *pvt)
{
    while(true) {
	pvt->sockfd = -1;
	pvt->connectClient();
	if(pvt->connected) {
            connectCallback(pvt->userPvt,tcpConnectYes);
            while(true) {
                int32 nnew = ::recv(pvt->sockfd,&pvt->inbuf[0],pvt->bufferSize,0);
                if(nnew<0) {
                    pvt->issueError("inettcpTask recv error");
                    break;
                }
                if(nnew==0) {
                    pvt->issueError("inettcpTask recv received 0 length message");
                    break;
                }
                if(inettcpDebug) pvt->dumpData("inettcpTask",pvt->inbuf,nnew);
                receiveCallback(pvt->userPvt,pvt->inbuf,nnew);
            }
        }
	pvt->connected = false;
	if(pvt->sockfd>0) {
	    int status = ::close(pvt->sockfd);
	    if(status) pvt->issueError("closing sockfd");
	}
        pvt->connectCallback(pvt->userPvt,tcpConnectNo);
	epicsThreadSleep(connectRetryDelay);
    }
}

extern "C" {
static const iocshArg initInetPortArg0 = { "portName",iocshArgString};
static const iocshArg initInetPortArg1 = { "inetAddr",iocshArgString};
static const iocshArg initInetPortArg2 = { "port",iocshArgInt};
static const iocshArg initInetPortArg3 = { "bufsize",iocshArgInt};
static const iocshArg * const initInetPortArgs[4] = {
    &initInetPortArg0,&initInetPortArg1,&initInetPortArg2,
    &initInetPortArg3};
static const iocshFuncDef initInetPortFuncDef = {
    "initInetPort",4,initInetPortArgs};
static void initInetPortCallFunc(const iocshArgBuf *args)
{
    initInetPort(args[0].sval,args[1].sval,args[2].ival,args[3].ival);
}

static void inetPortRegister(void)
{
    iocshRegister(&initInetPortFuncDef,initInetPortCallFunc);
}
epicsExportRegistrar(inetPortRegister);
} // end extern "C"
