//serialPortPeek.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/*
	Author: Marty Kraimer
	Date: 21OCT98

        Modifications:
        16-May-2003 Mark Rivers.  Added following features:
            - Output lines now contain the following information in order:
               - The name of the serial port.
               - The time in seconds that the first character on that line was
                 transmitted or received, in seconds after serialPortPeek was begun.
               - The character 'T' or 'R' indicating if the characters on that line
                 were Transmitted or Received.
               - The characters transmitted or received, up to a maximum of 20 per
                 line.  The characters are first printed in ASCII (if printable) and
                 then printed in hex.
            - A new line is printed whenever any of the following occurs:
               - A change from Transmit to Receive or from Receive to Transmit.
               - 20 characters have been received and are in the buffer.
               - More than 2 seconds of idle time have elapsed since the last time
                 a line was printed, and there are any characters in the buffer.
            - Increased the stack size from 2000 to 3000, it was marginal on PowerPC.

            These changes make serial "conversations" much easier to read.
            The timestamps are useful in analyzing performance. The idle timeout
            ensures that all characters sent or received are printed within 2 seconds.

*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <epicsRingBytes.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <cantProceed.h>
#include <iocsh.h>
#include <epicsExport.h>
 
#include "SerialPort.h"

class PeekData {
public:
    bool isReceive;
    unsigned char byte;
    double time;
};

#define RING_BUFF_SIZE 1000
#define MAX_CHARS_PER_LINE 20
#define MAX_IDLE_TIME 2.0
static int ringBufSize = RING_BUFF_SIZE * sizeof(PeekData);

class PeekPrivate {
public:
    epicsRingBytesId ringbuf;
    epicsTimeStamp startTime;
};

typedef struct peekInfo{
    epicsThreadId threadid;
    const char *portName;
    int nseconds;
}peekInfo;

static void peekThread(peekInfo *ppeekInfo);

static char threadname[] = "serialPeek";

extern "C" int serialPortPeek(const char *portName,int nseconds)
{
    peekInfo *ppeekInfo;

    ppeekInfo = (peekInfo *)callocMustSucceed(
        1,sizeof(peekInfo),"serialPortPeek");
    ppeekInfo->portName = portName;
    ppeekInfo->nseconds = nseconds;
    ppeekInfo->threadid = epicsThreadCreate(
        threadname,epicsThreadPriorityLow,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)peekThread,(void *)ppeekInfo);
    return(0);
}

static void peekHandler(void* parm,bool isReceive, unsigned char byte)
{
    PeekPrivate *pvt = (PeekPrivate *)parm;
    epicsTimeStamp timeNow;
    PeekData data;
    data.isReceive = isReceive;
    data.byte = byte;
    epicsTimeGetCurrent(&timeNow);
    data.time = epicsTimeDiffInSeconds(&timeNow,&pvt->startTime);
    epicsRingBytesPut(pvt->ringbuf,(char *)&data,sizeof(PeekData));
}
static void printQueue(const char *portName,PeekData data[],int len)
{
    int i;
    unsigned char byte;
    printf("%s %c %.2f ", portName, (data[0].isReceive ? 'R' : 'T'),
                          data[0].time);
    for(i=0;i<len; i++) {
        byte = data[i].byte;
        printf("%c", (isprint(byte) ? byte : '.'));
    }
    for(i=len;i<MAX_CHARS_PER_LINE+1; i++) printf("%c", ' ');
    for(i=0;i<len; i++) {
        byte = data[i].byte;
        printf("%2.2X ", byte);
    }
    printf("\n");
}
    
static void peekThread(peekInfo *ppeekInfo)
{
    const char *portName = ppeekInfo->portName;
    epicsTimeStamp startTime,currentTime,idleStart;
    double nseconds = ppeekInfo->nseconds;
    PeekPrivate pvt;
    PeekData data[MAX_CHARS_PER_LINE];
    int nextQueue = 0;
    bool isReceive = true;

    if((pvt.ringbuf = epicsRingBytesCreate(sizeof(void *) * ringBufSize))==NULL){
	printf("peekThread: initWatchDog failed\n");
        free(ppeekInfo);
	return;
    }
    if(!SerialPort::setPeekHandler(portName,&pvt,peekHandler)) return;
    epicsTimeGetCurrent(&startTime);
    pvt.startTime = startTime;
    epicsTimeGetCurrent(&idleStart);
    printf("\n");
    while(true) {
        double deltaTime;
        epicsThreadSleep(1.0);
        epicsTimeGetCurrent(&currentTime);
        deltaTime = epicsTimeDiffInSeconds(&currentTime,&startTime);
        if(epicsTimeDiffInSeconds(&currentTime,&idleStart)>=2.0 && nextQueue>0){
            printQueue(portName, data , nextQueue);
            epicsTimeGetCurrent(&currentTime);
            idleStart = currentTime;
            nextQueue = 0;
        }
	while (epicsRingBytesUsedBytes(pvt.ringbuf)>=(int)sizeof(PeekData)){
	    PeekData dataNow;
	    if(epicsRingBytesGet(pvt.ringbuf,(char *)&dataNow,
	    sizeof(PeekData))!=(int)sizeof(PeekData)) {
		printf("peekThread epicsRingBytesGet returned error\n");
	    } else {
                if(isReceive != dataNow.isReceive || nextQueue == MAX_CHARS_PER_LINE) {
                    printQueue(portName, data , nextQueue);
                    epicsTimeGetCurrent(&currentTime);
                    idleStart = currentTime;
		    nextQueue = 0;
                    isReceive = dataNow.isReceive;
                }
                data[nextQueue].isReceive = dataNow.isReceive;
                data[nextQueue].byte = dataNow.byte;
                data[nextQueue].time = dataNow.time;
                ++nextQueue;
	    }
            epicsTimeGetCurrent(&currentTime);
            if(deltaTime >= nseconds) break;
	}
        epicsTimeGetCurrent(&currentTime);
        if(deltaTime >= nseconds) break;
    }
    if(nextQueue>0) printQueue(portName, data , nextQueue);
    printf("serialPortPeek done\n");
    SerialPort::freePeekHandler(portName);
    epicsRingBytesDelete(pvt.ringbuf);
    free(ppeekInfo);
}

extern "C" {
static const iocshArg serialPortPeekArg0 = { "portName",iocshArgString};
static const iocshArg serialPortPeekArg1 = { "seconds",iocshArgInt};
static const iocshArg * const serialPortPeekArgs[2] = {
    &serialPortPeekArg0,&serialPortPeekArg1};
static const iocshFuncDef serialPortPeekFuncDef = {
    "serialPortPeek",2,serialPortPeekArgs};
static void serialPortPeekCallFunc(const iocshArgBuf *args)
{
    serialPortPeek(args[0].sval,args[1].ival);
}

static void serialPortPeekRegister(void)
{
    iocshRegister(&serialPortPeekFuncDef,serialPortPeekCallFunc);
}
epicsExportRegistrar(serialPortPeekRegister);
} // extern "C"
