// ttyVxPort.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/*
    Author: Marty Kraimer
    Date: 07JAN2003

*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ioLib.h>

#include <iocsh.h>
#include <epicsExport.h>
#include <epicsEvent.h>
#include <epicsThread.h>
#include <epicsTimer.h>
#include <epicsAssert.h>
#include <iocsh.h>
#include <mpfType.h>
#include "SerialPort.h"

extern "C" { 
int initTtyVxPort(const char *portName, const char *fileName,
    int baud, int bufsize);
static void timerCallback(void *ppvt);
static epicsTimerQueueId timerQueueId = 0;
int ttyVxPortDebug = 0;
epicsExportAddress(int, ttyVxPortDebug);
}


class TtyVxPort : public SerialPort{
public:
    TtyVxPort(const char *portName, const char *fileName,
        int baud, int bufsize);
    bool open();
    virtual serialStatus write(
	unsigned char *data, int nbytes, double timeout);
    virtual serialStatus read(double timeout);
    virtual void flush();
    virtual bool config( int baud, int stopBits, int bitsPerChar,
	char parity, char flowControl);
    bool config();
    const char *portName;
    const char *fileName;
    int baud;
    int fd;
    int bufsize;
    char *buffer;
    int bufcnt;  /* Number of characters in buffer. */
    fd_set readwritefs;
    int maxfd;
    epicsTimerId timerId;
};

static void timerCallback(void *ppvt)
{
    TtyVxPort *pTtyVxPort = (TtyVxPort *)ppvt;
    ioctl(pTtyVxPort->fd,FIOFLUSH,0);
    ioctl(pTtyVxPort->fd,FIOCANCEL,0);
}

int initTtyVxPort(const char *port, const char *file,
    int baud,int bufsize)
{
    char *portName,*fileName;
    errno = 0;
    if(!timerQueueId) {
        timerQueueId = epicsTimerQueueAllocate(true,epicsThreadPriorityLow);
    }
    int fd = ::open(file,O_RDWR|O_NOCTTY,0644);
    if(fd<0){
        printf("initTtyVxPort: open failed %s\n",strerror(errno));
        return(0);
    }
    /* Use lazy open in read/write. This makes all file access be done
     * In a single thread. This assumes that the server does read/write
     * form a single thread.*/
    close(fd);
    portName = (char *)calloc(strlen(port)+1,sizeof(char));
    strcpy(portName,port);
    fileName = (char *)calloc(strlen(file)+1,sizeof(char));
    strcpy(fileName,file);
    TtyVxPort *pTtyVxPort = new TtyVxPort(portName,fileName,baud,bufsize);
    pTtyVxPort->registerName(portName);
    printf("%s TtyVxPort WARNING "
           "can't set stopBits bitsPerChar parity flowControl\n",
        portName); 
    return(0);
}

TtyVxPort::TtyVxPort(const char *port, const char *file,
    int b, int bsize)
: portName(port), fileName(file), baud(b), fd(-1), bufsize(bsize), bufcnt(0)
{
    buffer = (char *)calloc(bufsize,sizeof(char));
    timerId = epicsTimerQueueCreateTimer(timerQueueId,timerCallback,this);
}

bool TtyVxPort::open()
{
    if(fd>=0) return(true);
    errno = 0;
    fd = ::open(fileName,O_RDWR|O_NOCTTY,0644);
    if(fd<0){
        printf("TtyVxPort:open failed %s\n",strerror(errno));
        return(false);
    }
    FD_ZERO(&readwritefs);
    FD_SET(fd,&readwritefs);
    maxfd = fd + 1;
    if(!config()) {
        printf("TtyVxPort::open %s config failed\n",fileName);
        return(false);
    }
    return(true);
}

bool TtyVxPort::config( int b, int sb, int bpc, char p, char fc)
{
    baud = b;
    if(fd<0 ) {return(open());}
    return(config());
}

bool TtyVxPort::config()
{
    int status;

    errno = 0;
    status = ioctl(fd,FIOBAUDRATE,baud);
    if(status) printf("%s TtyVxPort can't set baud %d error %s\n",
        portName,baud,strerror(errno)); 
    return(true);
}

serialStatus TtyVxPort::write(
    unsigned char *data, int nbytes, double timeOut)
{
    if(fd<0 ) {if(!open()) return(serialFailure);}
    errno = 0;
    int nout = ::write(fd,(char *)data,nbytes);
    if(nout!=nbytes) {
        printf("%s TtyVxPort::write nbytes %d but only %d written\n",
            fileName,nbytes,nout);
        return(serialFailure);
    }
    if(peekHandler) {
        for(int ind=0; ind<nout; ind++) {
            (*peekHandler)(peekHandlerParm,false,data[ind]);
        }
    }
    return(serialSuccess);
}

serialStatus TtyVxPort::read(double timeOut)
{
    struct timeval timeout;
    byteHandlerRC rc = byteHandlerError;

    if(fd<0 ) {if(!open()) return(serialFailure);}
    assert(byteHandler);
    errno = 0;
    fd_set temp_fd_set = readwritefs;
    timeout.tv_sec = (int)timeOut;
    timeout.tv_usec = (int)((timeOut - (double)(timeout.tv_sec))*1000000.0);
    epicsTimerStartDelay(timerId,timeOut);
    while(1) {
        int ind, nin;

        errno = 0;
        if(bufcnt==0)  bufcnt = ::read(fd,buffer,bufsize);
        nin = bufcnt;
        if(nin<0) {
            printf("TtyVxPort::read failure %s\n",strerror(errno));
            return(serialFailure);
        }
        if(nin==0) {
           if(ttyVxPortDebug)  printf("%s TtyVxPort::read timeOut\n",fileName);
            return(serialTimeout);
        }
        for(ind=0; ind<nin; ind++ ) {
            bufcnt--;
            if(peekHandler) (*peekHandler)
                (peekHandlerParm,true,(unsigned char)buffer[ind]);
            rc = (*byteHandler)(byteHandlerParm,(unsigned char)buffer[ind]);
            if(rc!=byteHandlerOK) {
                epicsTimerCancel(timerId);
                serialStatus ioStatus = (rc==byteHandlerEndRead)
                    ? serialSuccess : serialFailure;
                if(bufcnt>0) memmove(buffer, &buffer[ind+1], bufcnt);
                return(ioStatus);
            }
        }
    }
    return(serialSuccess);
}

void TtyVxPort::flush()
{
    bufcnt = 0;
    if(fd<0 ) {if(!open()) return;}
    int status = ioctl(fd,FIOFLUSH,0);
    if(status) printf("%s TtyVxPort can't flush error %s\n",
        portName,strerror(errno)); 
}

extern "C" {
static const iocshArg initTtyVxPortArg0 = { "portName",iocshArgString};
static const iocshArg initTtyVxPortArg1 = { "fileName",iocshArgString};
static const iocshArg initTtyVxPortArg2 = { "baud",iocshArgInt};
static const iocshArg initTtyVxPortArg3 = { "bufsize",iocshArgInt};
static const iocshArg * const initTtyVxPortArgs[8] = {
    &initTtyVxPortArg0,&initTtyVxPortArg1,&initTtyVxPortArg2,
    &initTtyVxPortArg3,};
static const iocshFuncDef initTtyVxPortFuncDef = {
    "initTtyVxPort",4,initTtyVxPortArgs};
static void initTtyVxPortCallFunc(const iocshArgBuf *args)
{
    initTtyVxPort(
        args[0].sval,args[1].sval,args[2].ival,args[3].ival);
}

static void ttyVxPortRegister(void)
{
    iocshRegister(&initTtyVxPortFuncDef,initTtyVxPortCallFunc);
}
epicsExportRegistrar(ttyVxPortRegister);
} // end extern "C"

