// clientChar8Array.cc
/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include <epicsEvent.h>
#include <epicsThread.h>
#include <cantProceed.h>

#include "Message.h"
#include "ConnectMessage.h"
#include "Char8ArrayMessage.h"
#include "DLList.h"
#include "test.h"

int clientChar8ArrayDebug = 0;
int clientChar8ArrayTimeout = 5;

typedef struct pvtData {
    connectStatus connectState;
    char8	valueSent[512];
    char8	valueReceived[512];
    epicsEventId sem;
}pvtData;

static void myCallback(Message *message,void *clientPvt)
{
    pvtData *ppvtData = (pvtData *)clientPvt;
    
    switch(message->getType()) {
    case messageTypeConnect:
	{
	    ConnectMessage *pConnectMessage = (ConnectMessage *)message;

	    printf("clientChar8Array messageTypeConnect status %s\n",
		connectStatusString[pConnectMessage->status]);
	    ppvtData->connectState = pConnectMessage->status;
	    delete pConnectMessage;
	    epicsEventSignal(ppvtData->sem);
	}
	break;
    case messageTypeChar8Array:
	{
            Char8ArrayMessage *pChar8ArrayMessage =
		(Char8ArrayMessage *)message;
	
	    ::strcpy(ppvtData->valueReceived,
		pChar8ArrayMessage->value);
	    if(clientChar8ArrayDebug>1) {
		char buffer[100];

		sprintf(buffer,"client value %s pChar8ArrayMessage %p\n",
		    ppvtData->valueReceived,(void *)pChar8ArrayMessage);
		printf("%s",buffer);
	    }
	    if(strcmp(ppvtData->valueReceived,ppvtData->valueSent)==0){
		if(clientChar8ArrayDebug>1) printf("epicsEventSignal\n");
		epicsEventSignal(ppvtData->sem);
	    }
	    delete pChar8ArrayMessage;
	}
	break;
    default:
	printf("got illegal response message\n");
	break;;
    }
}

typedef struct args {
    char *server;
    int slot;
    int32 clientChar8ArraySize;
    double delay;
}args;

static void clientChar8Array(void *inargs)
{
    args *pargs = (args *)inargs;
    const char *server = pargs->server;
    int slot = pargs->slot;
    int32 clientChar8ArraySize = pargs->clientChar8ArraySize;
    double delay = pargs->delay;

    MessageClient	*pMessageClient;
    pvtData		*ppvtData = new pvtData;
    int			status;
    Char8ArrayMessage 	*message;

    if(clientChar8ArraySize<=0) clientChar8ArraySize=16;
    ppvtData->connectState = connectNo;
    ppvtData->valueReceived[0] = 0;
    ppvtData->valueSent[0] = 0;
    pMessageClient = new MessageClient(myCallback,ppvtData);
    ppvtData->sem  = epicsEventMustCreate(epicsEventEmpty);
    status = pMessageClient->bind(server,slot);
    if(status) {
	printf("bind returned %d\n",status);
	return;
    }
    char8 nextValue = 'z';
    while(true) {
	if(ppvtData->connectState!=connectYes) {
	    epicsEventWaitWithTimeout(ppvtData->sem,1.0);
	    continue;
	}
	int nwait=0;
	while(strcmp(ppvtData->valueReceived,ppvtData->valueSent)!=0){
	    if(nwait==0)printf("clientChar8Array send %s receive %s\n",
		ppvtData->valueSent,ppvtData->valueReceived);
	    epicsEventWaitWithTimeout(ppvtData->sem,1.0);
	    if(nwait++ == 5) break;
	}
	if(strcmp(ppvtData->valueReceived,ppvtData->valueSent)==0){
	    nextValue += 1;
	    if(nextValue > 'z') nextValue = 'a';
	}
	message = new Char8ArrayMessage;
	message->allocValue(clientChar8ArraySize);
	message->setSize(clientChar8ArraySize);
	for(int i=0; i<(clientChar8ArraySize-1); i++)
	    message->value[i] = nextValue;
	message->value[clientChar8ArraySize-1] = 0;
	::strcpy(ppvtData->valueSent,message->value);
	if(clientChar8ArrayDebug) printf("sending %s\n",message->value);
	message->timeout=20;
	status = pMessageClient->send(message);
	if(status) {
	    printf("send returned %d\n",status);
	    if(delay==0.0) epicsThreadSleep(.01);
	} else {
	    epicsEventMustWait(ppvtData->sem);
	}
	if(delay>0.0) epicsThreadSleep(delay);
    }
}

extern "C" int initClientChar8Array(const char *server,int slot,
    int32 clientChar8ArraySize,double delay)
{
    args *pargs = (args *)callocMustSucceed(1,sizeof(args),"initClientChar8Array");
    epicsThreadId threadId;

    pargs->server = (char *)callocMustSucceed(strlen(server)+1,sizeof(char),"initClientChar8Array");
    strcpy(pargs->server,server);
    pargs->slot = slot;
    pargs->clientChar8ArraySize = clientChar8ArraySize;
    pargs->delay = delay;
    threadId = epicsThreadCreate("clientChar8Array",epicsThreadPriorityMedium,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)clientChar8Array,(void *)pargs);
    if(!threadId) printf("initClientChar8Array epicsThreadCreate Failure\n");
    return(0);
}
